#![allow(clippy::doc_markdown)]

#[macro_use]
extern crate nix;

#[macro_use]
extern crate serde_derive;
#[macro_use]
extern crate log;

#[cfg(test)]
#[macro_use]
extern crate error_chain;

#[cfg(test)]
#[macro_use]
extern crate proptest;

#[cfg(test)]
#[macro_use]
extern crate matches;

#[macro_use]
extern crate lazy_static;

pub mod engine;

#[cfg(feature = "dbus_enabled")]
pub mod dbus_api;

pub mod stratis;
